(* Copyright Univ. of Maryland, College Park and Microsoft Corporation, 2011.
 * Coco is distrubited under Microsoft Public License.
 * Authors: Nataliya Guts, Michael Hicks, Nikhil Swamy, Daan Leijen.  
*)

(*--------------------------------------------------------------*)
(* Part I. Header information                                   *)
(*                                                              *)
{

open Pars
open Lexing

(*let record_newline (lexbuf : lexbuf) = lexbuf.EndPos <- lexbuf.EndPos.NextLine*)
let incr_linenum lexbuf = 
    let pos = lexbuf.Lexing.lex_curr_p in
    lexbuf.Lexing.lex_curr_p <- { pos with
      Lexing.pos_lnum = pos.Lexing.pos_lnum + 1; 
      Lexing.pos_bol = pos.Lexing.pos_cnum; 
    }

}

(*--------------------------------------------------------------*)
(* Part II. Define some regular expressions                     *)
(*                                                              *)

let num        = ['0'-'9']+
let alpha      = ['a'-'z' 'A'-'Z']
let endident   = (alpha | num | ['_' '$'])*
let ident      = alpha+ endident
let var	       = '\'' ident
let upident    = ['A'-'Z'] endident
let character  = '\'' _ '\''
let monadname  = '\'' ident
let integer    = '-'? num
let floating   = num? '.' num
let whitespace = [' ' '\t' ]
let newline    = ('\n' | '\r' '\n')
let linecomment = '/' '/' [^'\n']*

(*--------------------------------------------------------------*)
(* Part III. Token generators and rules.                        *)
(*                                                              *)


rule token = parse
| whitespace { token lexbuf }                          (* keep lexing!*)
| newline    { incr_linenum lexbuf; token lexbuf }   (* record line number and keep lexing! *)
| "int"	     { TINT }
| "char"     { TCHAR }
| "float"    { TFLOAT }
| "unit"     { TUNIT }
| "bool"     { TBOOL }
| "string"   { TSTRING }
| "coercion" { TCOERCION }
| "monad"    { TMONAD }
| "("        { TLPAR }
| ")"        { TRPAR }
| "{"        { TCLPAR }
| "}"        { TCRPAR }
| ","        { TCOMMA }
| "<<"       { TLTLT }
| "<"        { TLT }
| ">"        { TGT }
| ":"        { TCOLON }
| "+"        { TPLUS }
| ident      { TIdentifier(lexeme lexbuf) }                     (* return a token carrying data!*)
| var        { TVarId(lexeme lexbuf) }                     (* return a token carrying data!*)
| integer    { TIntegerLiteral(int_of_string (lexeme lexbuf)) }    (* return a token carrying data!*)
| eof        { TEOF }                                           (* return a token for the EOF*)
| linecomment 
  	     { token lexbuf }
| "(*"       { comment lexbuf; token lexbuf }

     
and comment = parse
    | "(*"  { comment lexbuf; comment lexbuf }
    | "*)"  { () }
(*    | "\n"   { lexbuf.EndPos <- lexbuf.EndPos.NextLine;
               comment lexbuf } *)
    | eof    { failwith "Unterminated comment" }
    | _      { comment lexbuf }
