/* Copyright Univ. of Maryland, College Park and Microsoft Corporation, 2011.
 * Coco is distrubited under Microsoft Public License.
 * Authors: Nataliya Guts, Michael Hicks, Nikhil Swamy, Daan Leijen.  
*/

/*--------------------------------------------------------------*/
/* Part I. Header information                                   */
/*                                                              */
%{

open Lambd
open Typaux

%}
/*--------------------------------------------------------------*/
/* Part II. Token definitions                                   */
/*                                                              */

%start start
%token <string> TIdentifier
%token <string> TVarId
%token <int>    TIntegerLiteral

%token  
       TLPAR TRPAR TLTLT TCOMMA TLT TGT TCOLON
       TINT TBOOL TSTRING
       TCOERCION TCLPAR TCRPAR TPLUS
       TMONAD TFLOAT TUNIT TVALUNIT TCHAR TTYPE
       TEOF

/* Low precedence for constructs that grab indefinitely to the right
   (fun and if/then/else). */
%nonassoc TLT

/* This is the type of the data produced by a successful reduction of the 'start'*/
/* symbol:*/
%type < Lambd.monad list list * Typaux.signature > start

/*--------------------------------------------------------------*/
/* Part II. Token definitions                                   */
/*                                                              */

%%
start: 
   OptCoercionDefinitions 
   TEOF { $1}
;

OptCoercionDefinitions:
    { [],[] }
  | TCOERCION TCLPAR CoerDefList TCRPAR { [],List.rev $3 }
  | TCOERCION TCLPAR OrderingList CoerDefList TCRPAR { List.rev $3, List.rev $4 }

OrderingList:
  | Ordering { [List.rev $1]}
  | OrderingList TCOMMA Ordering { (List.rev $3) :: $1 }

Ordering:
  | Mbase TLTLT Mbase { let (m1,tl1),(m3,tl3) = $1,$3 in [MBase(m3,tl3);MBase(m1,tl1)] }
  | Ordering TLTLT Mbase { let (m,tl) = $3 in MBase(m,tl)::$1 }

CoerDefList:
  | CoerDef { [$1]}
  | CoerDefList TCOMMA CoerDef { $3::$1 }

CoerDef:
    | TMONAD TLPAR TIdentifier TCOMMA TIdentifier TCOMMA TIdentifier TRPAR 
      { Monad ($3,$5,$7,[]) }
    | TMONAD TLPAR TIdentifier TCOMMA TIdentifier TCOMMA TIdentifier TCOMMA TLT TypeList TGT TRPAR 
      { if List.length $10 > 0 && not !Util.indexed 
        then failwith "Option -indexed should be enabled to use indexed monads"
        else Monad ($3,$5,$7,$10) }
    | TIdentifier TCOLON Mbase TPLUS Mbase TLTLT Mbase 
      { let (b1,t1),(b2,t2),(b3,t3) = $3,$5,$7 in Morph(0,$1,b1,t1,b2,t2,b3,t3) }
    | TIdentifier TIntegerLiteral TCOLON Mbase TPLUS Mbase TLTLT Mbase 
      { let (b1,t1),(b2,t2),(b3,t3) = $4,$6,$8 in Morph($2,$1,b1,t1,b2,t2,b3,t3) }

Mbase:
    | TIdentifier TLT TypeList TGT { ($1,$3) }
    | TIdentifier { ($1,[]) }

Type:
    TBOOL { TBase("bool") }
  | TINT  { TBase("int") }
  | TFLOAT  { TBase("float") }
  | TUNIT {TBase("unit")}
  | TSTRING { TBase("string") }
  | TCHAR { TBase("char") }
  | TIdentifier { TBase($1) }
  | TVarId { TVar($1,ref None) }
  | TLPAR TypeList TRPAR TIdentifier { TApp(TBase $4,$2) }

TypeList:
  | Type MoreTypes { $1::$2 }
      
MoreTypes:
  | TCOMMA Type MoreTypes { $2::$3 }
  | { [] }
      
     
