(* Copyright Univ. of Maryland, College Park and Microsoft Corporation, 2011.
 * Coco is distrubited under Microsoft Public License.
 * Authors: Nataliya Guts, Michael Hicks, Nikhil Swamy, Daan Leijen.  
*)

module L = Lambd

open L
open Typaux
open Util

(************************************************************************)
(* Testing well-formedness *)

(* NG TODO both functions must be defined and can be given the monadic bind and unit types *)
let wfmonad env (m,fb,fu) = ()


(*****************************************)
(* TYPE AND CONSTRAINT UTILITY FUNCTIONS *)
(*****************************************)


let rec norm_typ t =
  match t with
    TVar (_,r) ->
      (match !r with
        None -> t
      | Some t -> ecr t)
  | TFun(t,m) -> TFun(norm_typ t, norm_mtyp m)
  | TTup(tl) -> TTup(List.map norm_typ tl)
  | TApp(t,tl) -> TApp(norm_typ t,List.map norm_typ tl)
  | _ -> t
and norm_mtyp (m,t) = 
  (mflatten m, norm_typ t)

let norm_ortyp = function
  | Pure t -> Pure (norm_typ t)
  | Monadic t -> Monadic (norm_mtyp t)


let fmvor = function
  | Pure p -> Lambd.fmv p
  | Monadic m -> Lambd.fmvm m


let tsubor t k t' = match t with
  | Pure t1 -> Pure (tsub t1 k t')
  | Monadic (mt2) -> Monadic (tsubm mt2 k t')

let tsubs t lkm =
  List.fold_left (fun t (k,m) -> tsub t k m) t lkm

let rec vars_of_mvar = function
  | MVar (c,r) -> (match !r with None -> [c] | Some r -> c :: vars_of_mvar r)
  | MBase (b,_) -> [b]

let submonad (x,nm) m =  
  if List.mem x (vars_of_mvar m) then nm else m
let rec msub t k m =
  let t' = ecr t in
  match t' with
    TVar _ 
  | TBase _ -> t'
  | TFun (t1,t2) -> TFun (msub t1 k m, msubm t2 k m)
  | TTup tl -> TTup (List.map (fun x -> msub x k m) tl)
  | TApp (t, tl) -> TApp (msub t k m, List.map (fun s -> msub s k m) tl)
  | TMorph (m1,m2) -> TMorph(msubm m1 k m,msubm m2 k m)
and msubm (m,t) k nm = 
	(submonad (k,nm) m),msub t k nm


let idv = function
    TVar(id,r) -> id
  | _ -> invalid_arg "expecting a TVar type"


let subm_cs cs subm =  
  let sub_monads m l = List.fold_left (fun t (k,m) -> submonad (k,m) t) m l in
  let csub (l,t1,t2) = (l, sub_monads  t1 subm,sub_monads  t2 subm) in
  let cssub (lst,t2) = (List.map csub lst), (sub_monads t2 subm) in
  List.map cssub cs

let instantiate (Scheme (tvs,cs,typ)) : tvar list *typ * coer_env =
  if Util.empty tvs then ([],typ,cs)
  else 
  (* fresh tvar for each bound variable *)
  let mvars, tvars = List.partition (fun x -> String.sub x 0 1 = "m") tvs in
  let subt = List.map (function tv -> (tv,freshtvar())) tvars in
  let subm,nvars = List.split (List.map (function tv -> let mname,m = freshmvar() in
                       if not !quiet then print_endline ("Inst: fresh "^mname^" for "^tv);
                       (tv,m),mname) mvars) in
  (* substitute all bound variables free in typ *)
  let typ' = List.fold_left (fun t (k,m) -> tsub t k m) typ subt in
  let typ'' = List.fold_left (fun t (k,m) -> msub t k m) typ' subm in
  (* TODO also refresh the constraint names? *)
  let cs' = subm_cs cs subm in
  (nvars,typ'',cs')

(* Given an environment, type, and constraint set, return
   a generalized type and the monad substitutions
   //was:any constraints that are not included in that generalized type. *)
let generalize (g:typ_env) (t:typ) (cs:coer_env) : scheme =
  let ttvs = ftv t in
  let gtvs = ftvg g in
  let ctvs = ftvcs cs in
  let tvs = setminus (union ctvs ttvs) gtvs in
  Scheme (tvs,cs,t)

let get_var = function (MVar(x,_)) -> x | _ -> failwith "Not a monad variable"

let mvars = function | MVar (_,_) as m -> [m]  | _ -> [] 

let abs_of_cstrs cs s = 
  let abs_of_constr s (lst,m2) (ml,cl) : var list * (var * typ option) list = 
    let abs_of_lst (l1,m1) (ml',cl') =
      let l1typ = "a"^l1 in
      let morphtyp = TMorph ((m1,TVar(l1typ, ref None)),(m2, TVar(l1typ, ref None))) in
      let ml'' = match mflatten m1 with 
       | MVar (mvar,_) -> if List.mem mvar ml' then ml' else mvar::ml'
       | _ -> ml' in
      ml'',(l1,Some morphtyp)::cl' in
    let ml',cl' = List.fold_right abs_of_lst lst (ml,cl) in
    let ml2' = match mflatten m2 with 
       | MVar (mvar,_) -> if List.mem mvar ml' then ml' else mvar::ml'
       | _ -> ml' in
    ml2',cl' in
  List.fold_right (abs_of_constr s) cs ([],[]) 

let ev_of_cstrs cs s = [],[]
(* TODO!!! 
  let ev_of_constr s (lst,m2) (ml,cl) : monad list * var list = 
    let of_lst (l1,m1) (ml',cl') =
      let ml'' = match mflatten m1 with 
       | MVar (mvar,_) -> if List.mem mvar (List.map get_var ml') then ml' else  m1::ml'
       | _ -> ml' in
      ml'',l1::cl' in
    let ml',cl' = List.fold_right of_lst lst (ml,cl) in
    let ml2' = match mflatten m2 with 
       | MVar (mvar,_) -> if List.mem mvar (List.map get_var ml') then ml' else m2::ml'
       | _ -> ml' in
    ml2',cl' in
  List.fold_right (ev_of_constr s) cs ([],[]) 
*)

let abstract_constraints (t:term) cs s: term = t
(* TODO!!! 
  if empty cs then t else
  let monads,morphs = abs_of_cstrs cs s in  
  EvAbs(monads,morphs,t)
*)

(*
let inst_evidence (t:ident) cs s: term = 
  if empty cs then Var t else
  let monads,mphs = ev_of_cstrs cs s in 
  (EvApp(t,(List.rev (monads)),List.rev mphs))
*)
let lookup_var (x:var) (g:typ_env) : scheme = 
  List.assoc x g

(********************)
(* TYPING ALGORITHM *)
(********************)

let is_value = function
  | L.Lam _ 
  | L.Rec _
  | L.Var _ 
  | L.Base _ -> true
  | _ -> false

let rec get_fun_args args = function 
  | App (e,arg) -> get_fun_args (arg::args) e
  | e -> e,List.rev args
let rec get_arg_types ts = function
  | TFun (targ,(MBase(s,[]),tres)) when s=bottom ->
      get_arg_types (targ::ts) tres
  | TFun (targ,(m,t)) ->
      m,t, List.rev (targ::ts)
  | t -> mBot,t,List.rev ts

let lift = ref false

let rec typing (s:signature) (g:typ_env) (e:term) : term * ortyp * constr list = 
  match e with
    Base (id,s) -> (e,Pure (TBase id),[])    
  | Var x -> 
    (try 
      let olds = lookup_var x g in
      let (nvars,t,cs) = instantiate olds in
      let e_res = if empty cs then Var x else
        let monads,mphs = ev_of_cstrs cs s in 
        let mds = List.fold_left (fun l x -> 
            if List.mem (get_var x) nvars then x::l else l) [] monads in
        Annot(EvApp(x,mds,List.rev (List.map (fun x -> Var x) mphs)),TLet olds) in
      (e_res,Pure t,cs)
    with Not_found -> failwith ("not_found "^x))
  | Lam (x,_,e1) -> 
      let tv = if x="()" then TBase("unit") else freshtvar () in
      let (e1',t,cs) = typing s ((x,Scheme ([],[],tv))::g) e1 in
      (match t with 
       | Pure p -> 
          (* To use TS-Lift, we inject a fresh monad with its unit;
             to use TS-Bot, we inject the Bot monad*)  
          if !lift then 
             let _,m = freshmvar () in
             (Lam(x,Some(tv),(e1')), Pure (TFun(tv, (m,p))), cs)
          else 
             (Lam(x,None,e1'), Pure (TFun(tv, (mBot,p))), cs)
       | Monadic tf -> (Lam (x,Some(tv),e1'),Pure (TFun(tv,tf)),cs))  
  | Rec (x,Annot(e1,TLet(annot))) ->
      let (e1',t,cs) = typing s ((x,annot)::g) e1 in
      let (_,t',_) = instantiate annot in
      (match t with 
       | Pure p -> unify p t'; (Rec(x,e1'),t,cs)
       | _ -> failwith ("monadic expression under recursion"))  
  | Rec (x,e1) ->
      let tv = freshtvar () in
      let (e1',t,cs) = typing s ((x,Scheme ([],[],tv))::g) e1 in
      (match t with 
       | Pure p -> unify p tv; (Rec(x,e1'),t,cs)
       | _ -> failwith ("monadic expression under recursion"))  

(* 
 | Tuple args ->  
      let mname,fmon = freshmvar() in
      let fbind,tlist,elist,cs = 
       List.fold_right (fun e (fbind,tlist,elist,cs) -> 
        let (e',t',cs') = typing s g e in
	(match t' with
        | Pure p -> 
	  (fbind, p::tlist,e'::elist,cs'@cs)
	| Monadic (m,p) -> 
          let nv = freshvar() in
	  let nl = freshvar() in
	  let cs'' = ([nl,m],fmon) :: (cs' @ cs) in
	  let fbind' = fun x -> Bind(fmon, (App (Var nl, e')), Lam (nv,None, fbind x)) in  
	  fbind', p::tlist, Var nv::elist, cs'' 
        )) args ((fun x -> x),[],[],[]) in
       if Util.empty cs
       then (Tuple args), Pure (TTup tlist), []
       else fbind (Unit (fmon,Tuple (elist))), Monadic (fmon, (TTup tlist)), cs 

 | App (Var id, Tuple args) when id = String.capitalize id -> 
 (* Constructor. Inline typing of Tuple to avoid an extra bind *)
      let (tvars,tname,targs) = lookup_cons s id in

      let freshtvars = List.map (fun x -> (x,freshtvar())) tvars in
      let freshtargs = List.fold_right (fun (k,m) res -> List.map (fun x -> tsub x k m) res) freshtvars targs in
      let e',t',cs = typing s g (Tuple args) in
      (match t' with
      | Pure (TTup tlist) -> 
          List.iter2 unify freshtargs tlist;
          App(Var id,e'), Pure (TApp(TBase tname,tlist)),cs
      | Monadic (m,TTup(tlist)) -> 
          List.iter2 unify freshtargs tlist;
          let mname,fmon = freshmvar() in
          let nv = freshvar() in
	  let nl = freshvar() in
	  let cs' = ([nl,m],fmon) :: cs in
         Bind(fmon,App (Var nl,e'),Lam(nv,None, Unit (fmon,App(Var id,Var nv)))), 
         Monadic(fmon,(TApp(TBase tname,tlist))),cs'

      | _ -> failwith "Arguments of a constructor should be a tuple")
*)
 | App (e1,e2) -> 
    let f,args = get_fun_args [] e in
    let frwr,(tresmon,tres,targs),csf = match typing s g f with
    | frwr,Pure tf,csf -> frwr,get_arg_types [] tf,csf
    | frwr,Monadic (m,t),_ -> frwr,(m,t,[]),[] in
    if false && List.length args > 1 && List.length targs > 1 then
      (* Special treatment for full applications of pure functions *)
      (* f e1 e2 --> bind e1 (fun x1 -> bind{m,bot,m} e2 (fun x2 -> f x1 x2)) *) 
      begin 
        let rewrite_arg ttmp ((fbind,elist,cstmp,args,bmon) as acc) = match args with
        | [] -> acc
        | arg::rargs ->
	   let _,fmon = freshmvar() in 
           let (e',t',cs') = typing s g arg in 
	   (match t' with
           | Pure p -> unify ttmp p; (fbind, e'::elist,cs'@cstmp,rargs,bmon)
	   | Monadic (m,p) -> 
             let nv = freshvar() in
	     let nl = freshvar() in
	     let cs'' = ([nl,m,bmon],fmon) :: (cs' @ cstmp) in
	  (*   let fbind' = fun x -> Bind(Var nl,(m,bmon,fmon), e', Lam (nv,None, fbind x)) in  
 	     fbind', Var nv::elist, cs'',rargs,fmon) in
      *)
             let fbind' = fun x -> Bind(Var nl,(m,bmon,fmon), e', fbind x) in  
 	     fbind', Var nv::elist, cs'',rargs,fmon) in 
    let (ftmp,argsrwr,csargs,argsrest,resmon) = 
        List.fold_right rewrite_arg targs ((fun x->x),[],[],args,tresmon) in
    let partial_app = ftmp frwr in
(*       
 let partial_app = ftmp (apply frwr argsrwr)in
*)
        let params_fun = partial_app,Monadic (resmon,tres), csf@csargs in
        List.fold_right (typing_app s g) argsrest params_fun 
      end
    else (* Usual function application *) 
      typing_app s g e2 (typing s g e1) 
  | Let(x,e1,e2) -> 
      	let (e1',ot1,cs1) = typing s g e1 in
	(match (norm_ortyp ot1) with
	| Pure t1 -> (* TS-LetGen *) 
            let e1'',sg1 = typing_let !Util.letgen s g x (e1',t1,cs1) in 
     	    let (e2',ot2,cs2) = typing s sg1 e2 in
            if not !quiet then (print_lambda (s,[]) e2'; Format.print_newline ());
            (*let bound = Annot(e1'',TLet s1) in *)
	    (Let(x,e1'',e2'),ot2,cs2) 
	| Monadic (m1,t1) -> (* TS-LetDo *)
          let (e2',ot2,cs2) = typing s ((x,Scheme([],[],t1))::g) e2 in
	  let _,nm = freshmvar() in
          let l1 = gen_lab() in
	  (match (norm_ortyp ot2) with
	    | Monadic (m2,t2) -> 
	      let cs = ([(l1,m1,m2)],nm) :: (cs1@cs2) in
              let app = Bind(Var l1,(m1,m2,nm),e1',Lam(x,Some t1,e2')) in
              app, Monadic(nm,t2),cs
(*
              let bound = Annot(app,TBind(nm,t1)) in
	      (Bind(nm,bound,Lam(x,Some t1,App(Var l2, e2'))), Monadic(nm,t2), cs)
*)	    | Pure t2 ->
	      let cs = ([(l1,m1,mBot)],nm) :: (cs1@cs2) in
              let app = Bind(Var l1,(m1,mBot,nm),e1',Lam(x,Some t1,e2')) in
              app, Monadic(nm,t2),cs
))
  | Seq (e1,e2) -> typing s g (Let("dummy",e1,e2))
(*     	let (e1',ot1,cs1) = typing s g e1 in
	let nt1 = norm_ortyp ot1 in
	(match nt1 with
        | Pure (t1) -> 
          (* begin match t1 with 
          | TBase("unit") -> () 
          | _ -> print_endline "Warning: the sequenced expression should have type unit"
          end; *)
          let (e2',ot2,cs2) = typing s g e2 in
          (Seq(e1',e2'),ot2,cs1@cs2)
	| Monadic (m1,t1) ->
          let (e2',ot2,cs2) = typing s g e2 in
          let x = "_" in
	  (match (norm_ortyp ot2) with
	    | Monadic (m2,t2) -> 
	      let mname,nm = freshmvar() in
              let l1,l2 = gen_lab(),gen_lab() in
	      let cs = ([(l1,m1);(l2,m2)],nm) :: (cs1@cs2) in
              let app = App(Var l1,e1') in
              let bound = Annot(app,TBind(nm,t1)) in
	      (Bind(nm,bound,Lam(x,None,App(Var l2, e2'))), Monadic(nm,t2), cs)
	    | Pure t2 ->
              let bound = Annot(e1',TBind(m1,t1)) in
	      (Bind(m1,bound,Lam(x,None, Unit(m1,e2'))), Monadic(m1,t2), (cs1@cs2))
))
*)
  | Annot(e,TBind(m,t)) -> 
        let (e',ot,cs) = typing s g e in
        let ltop = gen_lab() in
        (match ot with
        | Pure t' -> unify t t'; Unit(m,e'),Monadic(m,t),cs
        | Monadic (m',t') -> unify t t'; unify_monad m m';e',Monadic(m,t),cs) 
(*App(Var ltop,e'),Monadic(m,t),([ltop,m'],m)::cs) *)       

  | Cond(e1,e2,e3) -> 
        let tv = freshtvar () in
	let _,fmon = freshmvar() in
	let _,fmon' = freshmvar() in
	let _,mres = freshmvar() in
        let l1,l2,l3 = gen_lab(),gen_lab(),gen_lab() in
        let fe1,fe2,fe3 = freshvar(),freshvar(),freshvar() in
      	let (e1',ot1,cs1) = typing s g e1 in
      	let (e2',ot2,cs2) = typing s g e2 in
      	let (e3',ot3,cs3) = typing s g e3 in
        (match (ot1,ot2,ot3) with
        | Pure t1,Pure t2,Pure t3 -> 
           unify t1 (TBase "bool");
           unify tv t2; unify tv t3;
           (Cond(e1', e2', e3'),Pure tv, cs1@cs2@cs3)
        | Monadic (MBase (b,[]),t1),Pure t2,Pure t3 when b = bottom -> print_endline "Bot"; 
           unify t1 (TBase "bool");
           unify tv t2; unify tv t3;
           (Cond(e1', e2', e3'),Pure tv, cs1@cs2@cs3)
           
        | _ ->
	   let mc1,ct1= match ot1 with
           | Pure t1 -> mBot,t1
	   | Monadic (m1,t1) -> m1,t1
           in
	   let mbr1,tbr1 = match ot2 with
           | Pure t2 -> mBot,t2
	   | Monadic (m2,t2) -> m2,t2
           in
	   let mbr2,tbr2 = match ot3 with
           | Pure t3 -> mBot,t3
	   | Monadic (m3,t3) -> m3,t3
           in
           unify ct1 (TBase "bool");
           unify tv tbr1; unify tv tbr2;
           let cond c = Cond(c,Bind(Var l1,(mbr1,mBot,mres),e2',identity),Bind(Var l2,(mbr2,mBot,fmon),e3',identity)) in
           let cond_mon = Bind(Var l3,(mc1,fmon,mres),e1',Lam(fe1,Some(TBase("bool")), cond (Var fe1))) in
           let res,cs,monres = if (not !Util.debug) && mc1 = MBase("bot",[])
                  then cond e1', [([(l1,mbr1,mBot);(l2,mbr2,mBot)],fmon)],fmon 
                  else cond_mon,[([(l1,mbr1,mBot);(l2,mbr2,mBot)],fmon);([l3,mc1,fmon],mres)],mres in
           (res,Monadic(monres,tv),cs@cs1@cs2@cs3)
  )

 | _ -> raise (NYI "Typing")
and typing_let gen s g x (e1',t1,cs1)= 
      if not !quiet then 
       (print_endline ("Generalizing "^x);
        print_lambda (s,[]) e1'; Format.print_newline ();
        print_endline ("TYPE: "^string_of_typ t1));
      let unsolved, theta = Solver.solve ~where:"x" s cs1 (fmv t1) in
      let s1 = 
        if gen 
        then generalize g t1 unsolved 
        else Scheme([],unsolved,t1) in 
      let ne1 = List.fold_right (*subst_term*) rename_binds theta e1' in
      if not !quiet then 
       (print_endline ("AFTER solving");
        print_lambda (s,[]) ne1; Format.print_newline ();
        print_endline ("TYPE: "^string_of_scheme s1));
      let e1'' = abstract_constraints ne1 unsolved s in
      Annot(e1'',TLet s1), ((x,s1)::g)
  

and typing_app s g e2 (e1',ot1,cs1) = 
      let (e2',ot2,cs2) = typing s g e2 in
         let tva = freshtvar() in
         let tvb = freshtvar() in
	 let mname,tvm = freshmvar() in
	 let fmname,fmon = freshmvar() in 
         let l1,l2,l3 = gen_lab(),gen_lab(),gen_lab() in
         let f,arg = freshvar(),freshvar() in
         let fmonad,ftyp = match norm_ortyp ot1 with
           | Pure t1 ->(if !Util.debug then Some mBot else None),t1
           | Monadic (m1,t1) -> Some m1,t1 in
         unify ftyp (TFun (tva, (tvm,tvb)));
         let argmonad,argtyp = match norm_ortyp ot2 with
           | Pure t1 -> (if !Util.debug then Some mBot else None),t1
           | Monadic (m1,t1) -> Some m1,t1 in
         unify argtyp tva;
	 let _,tmpmon = freshmvar() in
         let c2,term2,farg,argres = match argmonad,tvm with 
            | None,MBase(b,[]) when b=bottom -> ([],e2',(fun x->x),mBot)
            | None, m -> [[l2,mBot,tvm],tmpmon], Var arg,(fun x-> Bind(Var l2,(mBot,tvm,tmpmon),e2', (*Lam(arg,None,x)*) x)),tmpmon
            | Some m,_ -> [[l2,m,tvm],tmpmon], Var arg,(fun x-> Bind(Var l2,(m,tvm,tmpmon),e2', (*Lam(arg,None,x)*) x)),tmpmon in
         let c1,term1,ffun,res = match fmonad with 
            | None -> [],e1',(fun x->x), argres
            | Some m -> [[l1,m,argres],fmon],Var f,(fun x->Bind(Var l1,(m,argres,fmon),e1',Lam(f,None,x))),fmon in
         let cs = c1@c2@cs1@cs2 in
         (* ffun (farg (App(term1,term2))), *)
         ffun (farg term1),
         Monadic (res,tvb), cs
(*
         let c1,ffun,fterm,ftyp = match norm_ortyp ot1 with
           | Pure t1 -> [],(fun x -> x),e1',t1
           | Monadic (m1,t1) -> 
              [(l1,m1)], (fun x-> Bind (fmon,(App(Var l1,e1')),Lam(f,Some t1,x))),Var f,t1 in
         begin try
         unify ftyp (TFun (tva, (tvm,tvb)));
         let c2,argfun,argterm,argtyp = match norm_ortyp ot2 with
           | Pure t2 -> [],(fun x -> x),e2',t2
           | Monadic (m2,t2) -> 
              if string_of_monad m2 = bottom then [],(fun x -> x),e2',t2 else
              [(l2,m2)], (fun x-> Bind (fmon,(App(Var l2,e2')),Lam(arg,Some t2,x))), Var arg,t2 in
         unify argtyp tva;
         let c3 = (l3,tvm) in
         let app = ffun (argfun (App(Var l3,App(fterm,argterm)))) in
         app,Monadic (fmon,tvb), ((c3::c1@c2),fmon)::cs1@cs2
   
      with Type_error e -> raise (Type_error e)
 *)     (* TODO!!!     let margname,marg = freshmvar() in
           unify ftyp (TMorph ((marg,tva), (tvm,tvb)));
           (* TODO check if e2 is an application before typing it? *)
           match norm_ortyp ot2 with
           | Pure _ -> failwith "Morphisms can only be applied to applications"
           | Monadic (m2,t2) -> 
              unify t2 tva; 
              unify_monad marg m2; 
              let c3 = (l3,tvm) in
              let app = ffun (App(Var l3,App(fterm,e2'))) in
              app,Monadic (fmon,tvb), (c3::c1,fmon)::cs1@cs2
         end
*)

let rec typing_str (s:signature) (g:typ_env) (si:stritem) : 
  stritem * typ_env * (term * ortyp * constr list) option = 
  match si with
  | StrEmpty -> StrEmpty,g,None
  | StrTerm e -> let e',t',cs' = typing s g e in StrEmpty, g, Some (e',t',cs')
  | StrLet (b,x,e1) -> begin 
         let tv = freshtvar () in
	  let (e1',ot1,cs1) = typing s ((x,Scheme ([],[],tv))::g) e1 in
	  match norm_ortyp ot1 with
	  | Pure t1 -> unify t1 tv;
            let e1'',sg1 = typing_let true s g x (e1',t1,cs1) in
	    StrLet (b,x,e1''),sg1,None 
          | Monadic (m,t) -> failwith "Toplevel bindings should only bind values"
    end
  | StrCons (s1,s2) -> 
    let s1',g1,None = typing_str s g s1 in
    let s2',g2,t2' = typing_str s g1 s2 in
    StrCons(s1',s2'),g2,t2'



let rec assoc x = function
  | [] -> failwith "Not_found"
  | (hd,r)::tl -> if List.mem (string_of_monad x) (vars_of_mvar hd) then r else assoc x tl 
let rec mem x = function
  | [] -> false
  | (hd)::tl -> if List.mem (string_of_monad x) (vars_of_mvar hd) then true else mem x tl 

let rec labmem (x,t) = function 
  | [] -> false
  | (hd,_)::tl -> if hd = x then true else labmem (x,t) tl 

let rec labappmem x = function 
  | [] -> false
  | (hd)::tl -> if hd = x then true else labappmem x tl 

let make_subset mem l1 l = 
  List.fold_right (fun x acc -> 
    if mem x l1 then (x :: acc) else acc) l []


  


let rec simplify (s:signature) (g:typ_env) (e:term) : term = 
  e
(*  match e with
  | Annot(EvApp (var,mds,newmorphs),TLet(Scheme(_,csold,_))) -> 
    let oldmds,oldmrphs = ev_of_cstrs csold s in
    if Util.empty oldmds
    then Var var
    else begin 
    let msub = List.combine oldmds (List.rev mds) in
    let csub = List.combine oldmrphs (List.rev newmorphs) in 
    let unsolved, _ = Solver.solve ~where:("Inst-"^var) s csold [] in
    let ms,labs = ev_of_cstrs unsolved s in
    let ms = List.filter (fun x -> match mflatten x with MVar _ -> true | _ -> false) ms in
    let sorted = make_subset mem ms oldmds in
    let newlabs = make_subset labappmem labs (oldmrphs) in 
    let nlabs = List.map (fun x -> List.assoc x csub) (newlabs) in 
    let nms = List.map (fun x -> assoc x msub) sorted in
    EvApp (var,List.rev nms,List.rev nlabs)
  end
  | Let(x,Annot(EvAbs (_,_,e1),TLet(Scheme(_,csold,_) as sc)),e2) ->
    let (_,t,csnew) = instantiate sc in
    let oldmds,oldmrphs = abs_of_cstrs csold s in
    let newms,newslabs = abs_of_cstrs csnew s in
    let msubs = List.combine newms oldmds in
    let msubs_ty = List.combine newms (List.map (fun x -> MVar(x,ref None)) oldmds) in
    let unsolved, theta = Solver.solve ~where:("Abs-"^x) s csnew [] in
    let ms,labs = abs_of_cstrs unsolved s in
    let sorted = make_subset List.mem ms newms in 
    let newlabs = List.map (fun (l,typ) -> 
         l, match typ with None -> None | Some ty -> Some (List.fold_left (fun t (k,m) -> msub t k m) ty msubs_ty)) labs in
    let newnewlabs = make_subset labmem newlabs oldmrphs in 
    let e1' = simplify s g e1 in
    let ne1 = List.fold_right subst_term theta e1' in
    let nms = List.map (fun (x:ident) -> (List.assoc x msubs)) sorted in 
    let ne1' = EvAbs(nms,newnewlabs,ne1)  in
    let ne2 = simplify s g e2 in
    Let(x,ne1',ne2)
  | EvAbs (a,b,e) -> EvAbs(a,b,simplify s g e)
  | Var _ | Base _ | EvApp _ -> e
  | Unit (m,e1) -> 
      let ne1 = simplify s g e1 in
      Unit(m,ne1)
  | App (e1,e2) -> 
      let ne1 = simplify s g e1 in
      let ne2 = simplify s g e2 in
      App(ne1,ne2)
  | Cond (e1,e2,e3) -> 
      let ne2 = simplify s g e2 in
      let ne3 = simplify s g e3 in
      Cond(e1,ne2,ne3)
  | Bind (m,e2,e3) -> 
      let ne2 = simplify s g e2 in
      let ne3 = simplify s g e3 in
      Bind (m,ne2,ne3) 
  | Let (x,e2,e3) -> 
      let ne2 = simplify s g e2 in
      let ne3 = simplify s g e3 in
      Let (x,ne2,ne3)
  | Seq (e2,e3) -> 
      let ne2 = simplify s g e2 in
      let ne3 = simplify s g e3 in
      Seq (ne2,ne3)
  | Annot(e1,t) ->
      let ne1 = simplify s g e1 in
      Annot(ne1,t)
  | Lam(x,t,e1) -> 
      let ne1 = simplify s g e1 in
      Lam(x,t,ne1)
  | Rec(x,e1) -> 
      let ne1 = simplify s g e1 in
      Rec(x,ne1)
  | Tuple(el) -> Tuple (List.map (simplify s g) el)
  | Pair (e1,e2) -> 
      let ne1 = simplify s g e1 in
      let ne2 = simplify s g e2 in
      Pair (ne1,ne2)
  | _ -> raise (NYI "simplify")
  
*)

